;----------------------------------------------------------------------
; Spannungswert (an Heileiter oder analogem Temperatursensor) ber ADC
; ermitteln und akustisch dezimal ausgeben. Auch verwendbar zur
; Ermittlung der Genauigkeit der internen Referenzspannung.
; Erluterungen und Makros: www.schramm-software.de/tipps
;----------------------------------------------------------------------
; Prozessor: ATtiny44
; Takt     : 1 MHz
; Sprache  : Assembler
; Version  : 1.0
; Autor    : Dr. Michael Schramm, siehe www.schramm-software.de
; Datum    : Februar 2012
;----------------------------------------------------------------------
; Portbelegung:
; PA0:     I  GND-Pegel aktiviert 1,1V-Referenz des ADCs
; PA4,PA5: O  Tonsignal (Piezo-Piepser oder Lautspr. ber R-C)
; PA6:     I  Analog-Eingang fr ADC-Messung (ADC6)
; PB0-PB2  O  VDD-Pegel, solange Messung aktiv (fr Sensor ber R1)
; PB3:     I  (Reset)
;----------------------------------------------------------------------
.equ F_CPU = 1000000
.include "tn44def.inc"
.include "../makros.inc"
; Als globale Variablen genutzte Register:
.def intreg  = r14 ;SREG-Zwischenspeicher bei Interrupt
.def timovl  = r15 ;Zhler fr Timer-Overflow-Interrupts
.def r_adcl  = r24 ;bergabe des ADC-Messwertes (Low-Byte)
.def r_adch  = r25 ;bergabe des ADC-Messwertes (High-Byte)
.def nullreg = r31 ;(ZH) Konstante Null
;----------------------------------------------------------------------
.DSEG ;SRAM-Belegung
dezzahl:  .byte 6
tmp_word: .byte 2
;----------------------------------------------------------------------
.CSEG
;Reset- und Interruptvektoren
  rjmp start     ; 1 POWER ON RESET
  reti           ; 2 0x0001 INT0 External Interrupt Request 0
  reti           ; 3 0x0002 PCINT0 Pin Change Interrupt Request 0
  reti           ; 4 0x0003 PCINT1 Pin Change Interrupt Request 1
  reti           ; 5 0x0004 WDT Watchdog Time-out
  reti           ; 6 0x0005 TIM1_CAPT  Timer/Counter1 Capture Event
  reti           ; 7 0x0006 TIM1_COMPA Timer/Counter1 Compare Match A
  reti           ; 8 0x0007 TIM1_COMPB Timer/Counter1 Compare Match B
  reti           ; 9 0x0008 TIM1_OVF Timer/Counter1 Overflow
  reti           ;10 0x0009 TIM0_COMPA Timer/Counter0 Compare Match A
  reti           ;11 0x000A TIM0_COMPB Timer/Counter0 Compare Match B
  rjmp i_t0ov    ;12 0x000B TIM0_OVF Timer/Counter0 Overflow
  reti           ;13 0x000C ANA_COMP Analog Comparator
  in r_adcl,ADCL ;14 0x000D ADC ADC Conversion Complete
  in r_adcl,ADCL ;(15 0x000E EE_RDY EEPROM Ready)
  in r_adch,ADCH ;(16 0x000F USI_STR USI START)
  reti           ;(17 0x0010 USI_OVF USI Overflow)
;----------------------------------------------------------------------
zehnerpot: ;Zehnerpotenzen als 2-Byte-Zahlen fr Funktion bin2dec
; Tabelle muss im Anfangsbereich des Speichers stehen, damit das
; High-Byte der Adresse 0 ist
.dw 10000,1000,100,10,1

start:
; zunchst Stromsparmanahmen und Initialisierung
  out_i DIDR0,0b11111110 ;PINA-Input-Buffer ausschalten
  out_i DDRA, 0b00110000 ;Datenrichtung von Port A
  out_i PORTA,0b00000001 ;Startzustand
  out_i DDRB, 0b0111     ;Datenrichtung von Port B
  out_i PORTB,0b1000     ;Startzustand (PB0 - PB2 zunchst auf GND)
  sbi ACSR,ACD           ;Analog-Comparator ausschalten
  ldi r18,6              ;Wert fr ADMUX (Input = ADC6 = PA6)
  sbic PINA,0            ;PA0 auf GND-Pegel?
  rjmp init2
  cbi PORTA,0            ;PA0 hochohmig, um Strom zu sparen
  sbr r18,(1<<REFS1)     ;spter interne ADC-Referenzspannung aktivieren
init2:
; ADC konfigurieren mit Interrupt und Vorteiler 8 (-> 125 kHz)
  out ADMUX,r18        ;Input = ADC6 = PA6
  out_i ADCSRA,(1<<ADIE)+(3<<ADPS0)
  outi_hl SP,RAMEND    ;Stackpointer setzen
  clr nullreg          ;bleibt dauerhaft null
  out_i TIMSK0,1<<TOIE0;Timer-Overflow-Interrupt aktivieren
  sei                  ;Interrupts generell zulassen
  ldi r17,8
  rcall waitr17        ;etwa 1/2 Sekunde warten
; Messung durchfhren
  out_i PORTB,0b111    ;VDD an PB0 - PB2, Strom an fr Messung
  sbi ADCSRA,ADEN      ;ADC einschalten (schaltet ggf. int. Ref.spg. ein)
  for r16,24,wait_ref_lp       ;gut 70 s warten, damit sich ggf. die
  next_down r16,wait_ref_lp    ;interne Referenzspannung stabilisiert
  out_i MCUCR,(1<<SE)+(1<<SM0) ;Noise-reduction-Modus
  sleep                        ;Warten auf ADC-Ergebnis (1. Messung)
; die erste Messung war nur zum 'Warmlaufen' des ADCs
  out MCUCR,r16                ;Noise-reduction-Modus
  sleep                        ;Warten auf ADC-Ergebnis (2. Messung)
  cbi ADCSRA,ADEN              ;ADC ausschalten
  out PORTB,nullreg            ;Stromfluss stoppen
; Messwert akustisch ausgeben
  ldi_hl x,24      ;Adresse von R24 (= R_ADCL)
  ldi_hl y,dezzahl ;Speicherbereich fr Dezimalzahl-Ergebnis
  rcall bin2dec
  tst r18          ;Sonderfall 0 bercksichtigen
  brne ziff_ausg
    st y,r18       ;Dezimalziffer 0 schreiben
    inc r18        ;1 Ziffer ausgeben, also '0'
  ziff_ausg:       ;R18 Ziffern ab (Y) ausgeben
    ld r0,y+
    rcall zifferton
  next_down r18,ziff_ausg
; Ausschalten
  out_i MCUCR,(1<<SE)+(1<<SM1) ;Power-down-Modus
  sleep                        ;Tiefschlaf, warten auf Reset

; ******************* Unterprogramme / Funktionen *********************

tonsignal: ; *** Tonsignal an PB0/PB1 ausgeben
; (schlichte Realisierung ber Zhlschleife)
; R16: Periodenlnge als Vielfaches von 100 s
; R17: 1/10 der Anzahl der zu erzeugenden Perioden
  push_hl x
  push r20
  push r2
  sbi PORTA,5              ;Startzustand der Ports A4 und A5
  tonsiglp:
    for r20,10,ton_10per   ;10 Perioden erzeugen
      for xh,2,ton_period  ;fr die beiden halben Perioden
        mov r2,r16         ;R16 halbe Perioden erzeugen
        wt_halb_per:       ;1/2 Periode erzeugen
          for xl,16,wt50lp ;50 s warten (bei Taktfrequenz 1 MHz)
          next_down xl,wt50lp
        next_down r2,wt_halb_per
        ldi xl,3<<4
        out PINA,xl        ;Ports A4 und A5 umschalten
      next_down xh,ton_period
    next_down r20,ton_10per
  next_down r17,tonsiglp
  cbi PORTA,4              ;B0 und B1 auf 0 (spart ggf.
  cbi PORTA,5              ;Energie am Lautspr.)
  pop r2
  pop r20
  pop_hl x
ret
;----------------------------------------------------------------------
zifferton: ; *** R0 als Tonsignalfolge ausgeben
  push r0
  push r16
  push r17
  tst r0
  brne ziff_pieps
  ldi r16,20         ;500 Hz fr die Ziffer 0
  ldi r17,25         ;250 Perioden = 1/2 Sekunde
  inc r0             ;damit die folgende Schleife 1mal durchlaufen wird
  rjmp ausg_ziff_ton ;Sprung in die Schleife ...
  ziff_pieps:        ;R0 Piepser fr eine Dezimalziffer
    ldi r16,10       ;1 kHz
    ldi r17,25       ;250 Perioden = 1/4 Sekunde
    ausg_ziff_ton:
    rcall tonsignal
    ldi r17,4
    rcall waitr17    ;nach jedem Pieps 4/15 Sek. warten (knapp 270 ms)
  next_down r0,ziff_pieps
  ldi r17,10
  rcall waitr17      ;nach jeder Ziffer zustzlich gut 660 ms Pause
  pop r17
  pop r16
  pop r0
ret
;----------------------------------------------------------------------
; Warte-Routinen sind energiesparend mit Timer + SLEEP-Modus realisiert
waitr17: ;  *** R17 * 1/15 Sekunde warten (ca.)
  push r16
  clr timovl ;berlauf-Reg. fr den Timer, wird in Int.-Rout. inkrementiert
  out TCNT0,nullreg       ;Counter-Wert nullsetzen
  out_i TCCR0B,(1<<CS02)  ;Timer starten mit Systemtakt / 256 = 3,90625 kHz
  wtr17lp:
    out_i MCUCR,(1<<SE)   ;Idle-Modus
    sleep                 ;CPU schlft, wartet auf Interrupt (Timer)
	cp timovl,r17         ;r17 Timer-berlufe erreicht?
  brcs wtr17lp            ;noch nicht => weiter warten
  out TCCR0B,nullreg      ;Timer stoppen
  pop r16
ret
;----------------------------------------------------------------------
bin2dec: ; *** Dezimaldarstellung einer 2-Byte-Binrzahl berechnen
; Input:  X zeigt auf Beginn der Binrzahl (LSB) im SRAM
;         Y zeigt auf Beginn des SRAM-Speicherbereichs der Dezimalzahl
; Output: Dezimalzahl ab (Y), eine Dez.ziffer pro Byte, Start mit MSD
;         R18 = Anzahl der Dezimalziffern (0 bei 0)
; Binrzahl (X) wird zu null abgebaut
  push r0
  push r1
  push r10
  push r11
  push r16
  push r17
  push_hl z
  clr r18
  mov r11,xl             ;R11 = Kopie von XL
  ldi_hl z,(zehnerpot*2) ;mal 2 wegen wortweiser Adressierung im Flash
  for r16,5,b2d_nxt_ziff ;R16 = Abwrtzzhler fr Zehnerpotenzen
    clr r10              ;R10 = aktuelle Ziffer
    push yl
    ldi yl,low(tmp_word)
; hier wird ausgenutzt, dass Yh im kleinen SRAM immer gleich bleibt (0)
    for r17,2,b2d_pot_copy
      lpm r0,z+          ;Zehnerpotenz vom Flash ins SRAM kopieren
      st y+,r0
    next_down r17,b2d_pot_copy
    mov xl,r11
    adiw x,2
   b2d_nxt_test:
    for r17,2,b2d_compare
      ld r0,-y           ;Zehnerpotenz
      ld r1,-x           ;Rest der Ausgangszahl
      cp r1,r0
      brcs b2d_ziff_ok   ;kein weiterer Abzug mglich
      brne b2d_subtract
    next_down r17,b2d_compare
   b2d_subtract:
    mov xl,r11
    ldi yl,low(tmp_word)
    rcall x_min_y
    inc r10
    rjmp b2d_nxt_test
   b2d_ziff_ok:
    pop yl
    tst r18
    brne b2d_st_ziff
    tst r10
    breq b2d_ziff_fertig
    b2d_st_ziff:
    st y+,r10
    inc r18
  b2d_ziff_fertig:
  next_down r16,b2d_nxt_ziff
  sub yl,r18             ;Y wieder auf Dezimalzahlbeginn
  pop_hl z
  pop r17
  pop r16
  pop r11
  pop r10
  pop r1
  pop r0
ret
;----------------------------------------------------------------------
x_min_y: ; *** Subtraktion, 2-Byte-Integer (X) <- (X) - (Y)
; r0, r1 werden verndert
; x und y zeigen anschlieend hinter die Zahlen, sind also um 2 erhht
  push r17
  clc
  for r17,2,xminy_lp
    ld r0,x
    ld r1,y+
    sbc r0,r1
    st x+,r0
  next_down r17,xminy_lp
; sbiw x,2
; sbiw y,2
  pop r17
ret

; ************************ Interrupt-Routinen *************************

i_t0ov: ; *** Timer/Counter Overflow
; Prescaler=256 => 18,3 Overflow-Interrupts pro Sekunde
  in intreg,sreg  ;Statusregister merken
  inc timovl      ;Zhler fr (in etwa) 18tel Sekunden
  out sreg,intreg ;Statusregister wiederherstellen
reti

; ******************************* ENDE *******************************
